from typing import Callable

from gi.repository import Adw, Gtk

from cozy.control.artwork_cache import ArtworkCache
from cozy.ext import inject
from cozy.model.book import Book

BOOK_ICON_SIZE = 52


class BookRow(Adw.ActionRow):
    _artwork_cache: ArtworkCache = inject.attr(ArtworkCache)

    def __init__(
        self, book: Book, on_click: Callable[[Book], None] | None = None
    ) -> None:
        super().__init__(
            title=book.name, subtitle=book.author, selectable=False, use_markup=False
        )

        if on_click is not None:
            self.connect("activated", lambda *_: on_click(book))
            self.set_activatable(True)
            self.set_tooltip_text(_("Play this book"))

        paintable = self._artwork_cache.get_cover_paintable(
            book, self.get_scale_factor(), BOOK_ICON_SIZE
        )
        if paintable:
            album_art = Gtk.Picture.new_for_paintable(paintable)
            album_art.add_css_class("round-6")
            album_art.set_overflow(True)
        else:
            album_art = Gtk.Image.new_from_icon_name("book-open-variant-symbolic")
            album_art.set_pixel_size(BOOK_ICON_SIZE)

        album_art.set_size_request(BOOK_ICON_SIZE, BOOK_ICON_SIZE)
        album_art.set_margin_top(6)
        album_art.set_margin_bottom(6)

        self.add_prefix(album_art)



import logging
import time
from threading import Event, Thread
from typing import Optional, Callable

import gi

from cozy.control.artwork_cache import ArtworkCache
from cozy.ext import inject
from cozy.model.book import Book
from cozy.model.chapter import Chapter
from cozy.report import reporter
from cozy.ui.chapter_element import ChapterElement
from cozy.ui.disk_element import DiskElement
from cozy.view_model.book_detail_view_model import BookDetailViewModel

from gi.repository import Adw, GLib, Gtk

log = logging.getLogger("BookDetailView")


ALBUM_ART_SIZE = 256


@Gtk.Template.from_resource('/com/github/geigi/cozy/book_detail.ui')
class BookDetailView(Gtk.Box):
    __gtype_name__ = 'BookDetail'

    play_book_button: Gtk.Button = Gtk.Template.Child()

    book_label: Gtk.Label = Gtk.Template.Child()
    author_label: Gtk.Label = Gtk.Template.Child()
    last_played_label: Gtk.Label = Gtk.Template.Child()
    total_label: Gtk.Label = Gtk.Template.Child()

    remaining_label: Gtk.Label = Gtk.Template.Child()
    book_progress_bar: Gtk.ProgressBar = Gtk.Template.Child()

    published_label: Gtk.Label = Gtk.Template.Child()
    published_text: Gtk.Label = Gtk.Template.Child()

    download_box: Gtk.Box = Gtk.Template.Child()
    download_label: Gtk.Label = Gtk.Template.Child()
    download_image: Gtk.Image = Gtk.Template.Child()
    download_switch: Gtk.Switch = Gtk.Template.Child()

    album_art: Gtk.Picture = Gtk.Template.Child()
    album_art_container: Gtk.Box = Gtk.Template.Child()

    unavailable_box: Gtk.Box = Gtk.Template.Child()

    chapters_stack: Gtk.Stack = Gtk.Template.Child()
    chapter_box: Gtk.Box = Gtk.Template.Child()
    book_overview_scroller: Gtk.ScrolledWindow = Gtk.Template.Child()

    main_flow_box: Gtk.FlowBox = Gtk.Template.Child()

    _view_model: BookDetailViewModel = inject.attr(BookDetailViewModel)
    _artwork_cache: ArtworkCache = inject.attr(ArtworkCache)

    _current_selected_chapter: Optional[ChapterElement] = None

    def __init__(self, main_window_builder: Gtk.Builder):
        super().__init__()

        self._navigation_view: Adw.NavigationView = main_window_builder.get_object("navigation_view")
        self._book_details_container: Adw.ToolbarView = main_window_builder.get_object("book_details_container")
        self._book_details_container.set_content(self)

        headerbar = Adw.HeaderBar()
        self.header_title = Adw.WindowTitle()
        headerbar.set_title_widget(self.header_title)
        self._book_details_container.add_top_bar(headerbar)

        self.book_overview_scroller.props.propagate_natural_height = True

        self._chapters_event: Event = Event()
        self._chapters_thread: Thread = None
        self._prepare_chapters_job()

        self._connect_view_model()
        self._connect_widgets()

    def _connect_view_model(self):
        self._view_model.bind_to("book", self._on_book_changed)
        self._view_model.bind_to("playing", self._on_play_changed)
        self._view_model.bind_to("is_book_available", self._on_book_available_changed)
        self._view_model.bind_to("downloaded", self._set_book_download_status)
        self._view_model.bind_to("current_chapter", self._on_current_chapter_changed)
        self._view_model.bind_to("last_played_text", self._on_last_played_text_changed)
        self._view_model.bind_to("remaining_text", self._on_times_changed)
        self._view_model.bind_to("progress_percent", self._on_times_changed)
        self._view_model.bind_to("total_text", self._on_times_changed)
        self._view_model.bind_to("playback_speed", self._on_times_changed)
        self._view_model.bind_to("lock_ui", self._on_lock_ui_changed)
        self._view_model.bind_to("open", self._on_open)

    def _connect_widgets(self):
        self.play_book_button.connect("clicked", self._play_book_clicked)
        self.download_switch.connect("state-set", self._download_switch_changed)

    def _on_book_changed(self):
        if not self._view_model.book:
            msg = "ViewModel book was None."
            log.warning(msg)
            reporter.warning("BookDetailView", msg)
            return

        self._chapters_event.clear()

        book = self._view_model.book

        self.chapters_stack.set_visible_child_name("chapters_loader")
        self.book_overview_scroller.set_visible(False)
        self._run_display_chapters_job(book)

        self._current_selected_chapter = None

        self.published_label.set_visible(False)
        self.published_text.set_visible(False)
        self.total_label.set_visible(False)
        self.unavailable_box.set_visible(False)

        self.book_label.set_text(book.name)
        self.author_label.set_text(book.author)
        self.header_title.set_title(book.name)
        self.header_title.set_subtitle(book.author)

        self.last_played_label.set_text(self._view_model.last_played_text)

        self._set_cover_image(book)

        self._display_external_section()
        self._set_progress()

    def _on_open(self):
        if self._navigation_view.props.visible_page.props.tag == "book_overview":
            self._navigation_view.pop_to_tag("book_overview")
        else:
            self._navigation_view.push_by_tag("book_overview")

    def _on_play_changed(self):
        playing = self._view_model.playing

        if playing:
            self.play_book_button.set_icon_name("media-playback-pause-symbolic")
        else:
            self.play_book_button.set_icon_name("media-playback-start-symbolic")

        if self._current_selected_chapter:
            self._current_selected_chapter.set_playing(playing)
        else:
            log.error("_current_selected_chapter is null. Skipping...")
            reporter.error("book_detail_view",
                           "_current_selected_chapter was NULL. No play/pause chapter icon was changed")

    def _on_book_available_changed(self):
        info_visibility = not self._view_model.is_book_available
        self.unavailable_box.set_visible(info_visibility)

    def _on_current_chapter_changed(self):
        if self._current_selected_chapter:
            self._current_selected_chapter.deselect()
            self._current_selected_chapter.set_playing(False)

        current_chapter = self._view_model.current_chapter

        for child in self.chapter_box:
            if not isinstance(child, ChapterElement):
                continue

            if child.chapter == current_chapter:
                self._current_selected_chapter = child
                child.select()
                child.set_playing(self._view_model.playing)
                break

    def _on_last_played_text_changed(self):
        self.last_played_label.set_text(self._view_model.last_played_text)

    def _on_times_changed(self):
        self.total_label.set_text(self._view_model.total_text)
        self._set_progress()

    def _on_lock_ui_changed(self):
        lock = self._view_model.lock_ui
        self.download_switch.set_sensitive(not lock)

    def _run_display_chapters_job(self, book):
        self._chapters_event.clear()
        # The job might be running on another thread. Attempt to cancel it first, wait a while and trigger the new one.
        self._interrupt_chapters_jobs()
        time.sleep(0.05)
        # This is done on a the UI thread to prevent chapters from the previous book flashing before the new chapters
        # are ready
        self._schedule_chapters_clearing()
        self._prepare_chapters_job()
        self._chapters_thread: Thread = Thread(target=self._schedule_chapters_rendering,
                                               args=[book, self._on_chapters_displayed])
        self._chapters_thread.start()

    def _schedule_chapters_rendering(self, book: Book, callback: Callable):
        disk_number = -1
        multiple_disks = self._view_model.disk_count > 1

        for chapter in book.chapters:
            if self._chapters_job_locked:
                self._schedule_chapters_clearing()
                return

            if multiple_disks and disk_number != chapter.disk:
                GLib.MainContext.default().invoke_full(GLib.PRIORITY_DEFAULT_IDLE, self._add_disk, book.id, chapter)

            GLib.MainContext.default().invoke_full(GLib.PRIORITY_DEFAULT_IDLE, self._add_chapter, book.id, chapter)

            disk_number = chapter.disk

            # TODO We need a timeout value
            self._chapters_event.wait()
            self._chapters_event.clear()

        GLib.MainContext.default().invoke_full(GLib.PRIORITY_DEFAULT_IDLE, callback)

    def _on_chapters_displayed(self):
        self.total_label.set_text(self._view_model.total_text)
        self.total_label.set_visible(True)
        self._set_book_download_status()

        self._on_current_chapter_changed()
        self._on_play_changed()
        self._on_book_available_changed()

        self.book_overview_scroller.set_visible(True)
        self.chapters_stack.set_visible_child_name("chapters_wrapper")

    def _display_external_section(self):
        external = self._view_model.is_book_external
        self.download_box.set_visible(external)
        self.download_switch.set_visible(external)

        if external:
            self.download_switch.handler_block_by_func(self._download_switch_changed)
            self.download_switch.set_active(self._view_model.book.offline)
            self.download_switch.handler_unblock_by_func(self._download_switch_changed)

    def _add_disk(self, book_id: int, chapter: Chapter):
        if book_id != self._view_model.book.id:
            return

        disc_element = DiskElement(chapter.disk)
        self.chapter_box.append(disc_element)
        self._chapters_event.set()

    def _add_chapter(self, book_id: int, chapter: Chapter):
        if book_id != self._view_model.book.id:
            return

        chapter_element = ChapterElement(chapter)
        chapter_element.connect("play-pause-clicked", self._play_chapter_clicked)
        self.chapter_box.append(chapter_element)
        self._chapters_event.set()

    def _schedule_chapters_clearing(self):
        GLib.MainContext.default().invoke_full(GLib.PRIORITY_DEFAULT_IDLE, self.chapter_box.remove_all_children)

    def _set_progress(self):
        self.remaining_label.set_text(self._view_model.remaining_text)
        self.book_progress_bar.set_fraction(self._view_model.progress_percent)

    def _set_cover_image(self, book: Book):
        paintable = self._artwork_cache.get_cover_paintable(book, self.get_scale_factor(), ALBUM_ART_SIZE)
        if paintable:
            self.album_art_container.set_visible(True)
            self.album_art.set_paintable(paintable)
            self.album_art.set_overflow(True)
        else:
            self.album_art_container.set_visible(False)

    def _interrupt_chapters_jobs(self):
        self._chapters_job_locked = True

        try:
            self._chapters_thread.join(timeout=0.2)
        except AttributeError as e:
            pass

    def _prepare_chapters_job(self):
        self._chapters_job_locked: bool = False

    def _download_switch_changed(self, _, state: bool):
        self._view_model.download_book(state)

    def _set_book_download_status(self):
        if not self._view_model.is_book_external:
            return

        if self._view_model.book.downloaded:
            icon_name = "downloaded-symbolic"
            text = _("Downloaded")
        else:
            icon_name = "download-symbolic"
            text = _("Download")

        self.download_image.set_from_icon_name(icon_name)
        self.download_label.set_text(text)

    def _play_chapter_clicked(self, _, chapter: Chapter):
        self._view_model.play_chapter(chapter)

    def _play_book_clicked(self, _):
        self._view_model.play_book()




from typing import Any
from gi.repository import Adw, Gio, Gtk

from cozy.ext import inject
from cozy.ui.widgets.error_reporting import ErrorReporting

from cozy.ui.widgets.storages import StorageLocations
from cozy.view_model.settings_view_model import SettingsViewModel


@Gtk.Template.from_resource("/com/github/geigi/cozy/preferences.ui")
class PreferencesView(Adw.PreferencesWindow):
    __gtype_name__ = "PreferencesWindow"

    _glib_settings: Gio.Settings = inject.attr(Gio.Settings)
    _view_model: SettingsViewModel = inject.attr(SettingsViewModel)

    storages_page: Adw.PreferencesPage = Gtk.Template.Child()
    user_feedback_preference_group: Adw.PreferencesGroup = Gtk.Template.Child()

    dark_mode_switch: Adw.SwitchRow = Gtk.Template.Child()
    swap_author_reader_switch: Adw.SwitchRow = Gtk.Template.Child()
    replay_switch: Adw.SwitchRow = Gtk.Template.Child()
    sleep_timer_fadeout_switch: Adw.SwitchRow = Gtk.Template.Child()
    artwork_prefer_external_switch: Adw.SwitchRow = Gtk.Template.Child()

    rewind_duration_adjustment: Gtk.Adjustment = Gtk.Template.Child()
    forward_duration_adjustment: Gtk.Adjustment = Gtk.Template.Child()
    fadeout_duration_adjustment: Gtk.Adjustment = Gtk.Template.Child()

    def __init__(self, **kwargs: Any) -> None:
        main_window = inject.instance("MainWindow")
        super().__init__(transient_for=main_window.window, **kwargs)

        error_reporting = ErrorReporting()
        error_reporting.show_header(False)
        self.user_feedback_preference_group.add(error_reporting)

        self.storage_locations_view = StorageLocations()
        self.storages_page.add(self.storage_locations_view)

        self._view_model.bind_to("lock_ui", self._on_lock_ui_changed)
        self._bind_settings()

    def _bind_settings(self) -> None:
        self._glib_settings.bind(
            "dark-mode", self.dark_mode_switch, "active", Gio.SettingsBindFlags.DEFAULT
        )

        self._glib_settings.bind(
            "swap-author-reader",
            self.swap_author_reader_switch,
            "active",
            Gio.SettingsBindFlags.DEFAULT,
        )

        self._glib_settings.bind(
            "replay", self.replay_switch, "active", Gio.SettingsBindFlags.DEFAULT
        )
        self._glib_settings.bind(
            "rewind-duration",
            self.rewind_duration_adjustment,
            "value",
            Gio.SettingsBindFlags.DEFAULT,
        )
        self._glib_settings.bind(
            "forward-duration",
            self.forward_duration_adjustment,
            "value",
            Gio.SettingsBindFlags.DEFAULT,
        )

        self._glib_settings.bind(
            "sleep-timer-fadeout",
            self.sleep_timer_fadeout_switch,
            "enable-expansion",
            Gio.SettingsBindFlags.DEFAULT,
        )

        self._glib_settings.bind(
            "sleep-timer-fadeout-duration",
            self.fadeout_duration_adjustment,
            "value",
            Gio.SettingsBindFlags.DEFAULT,
        )

        self._glib_settings.bind(
            "prefer-external-cover",
            self.artwork_prefer_external_switch,
            "active",
            Gio.SettingsBindFlags.DEFAULT,
        )

    def _on_lock_ui_changed(self) -> None:
        self.storage_locations_view.set_sensitive(not self._view_model.lock_ui)



import logging
import os
import platform
import sys
import threading
from pathlib import Path
from traceback import format_exception

import distro
from gi.repository import Adw, GLib

from cozy.app_controller import AppController
from cozy.control.db import init_db
from cozy.control.mpris import MPRIS
from cozy.db.settings import Settings
from cozy.db.storage import Storage
from cozy.report import reporter
from cozy.ui.main_view import CozyUI
from cozy.ui.widgets.filter_list_box import FilterListBox
from cozy.version import __version__

log = logging.getLogger("application")


class Application(Adw.Application):
    ui: CozyUI
    app_controller: AppController

    def __init__(self, pkgdatadir: str):
        self.pkgdatadir = pkgdatadir

        super().__init__(application_id='com.github.geigi.cozy')
        self.init_custom_widgets()

        GLib.setenv("PULSE_PROP_media.role", "music", True)
        GLib.set_application_name("Cozy")

        threading.excepthook = self.handle_exception

    def do_startup(self):
        log.info(distro.linux_distribution(full_distribution_name=False))
        log.info("Starting up cozy %s", __version__)
        log.info("libadwaita version: %s", Adw._version)

        self.ui = CozyUI(self.pkgdatadir, self, __version__)
        Adw.Application.do_startup(self)
        init_db()
        self.ui.startup()

    def do_activate(self):
        main_window_builder = self.ui.get_builder()
        self.app_controller = AppController(self, main_window_builder, self.ui)

        self.ui.activate(self.app_controller.library_view)

        if Settings.get().first_start:
            Settings.update(first_start=False).execute()

            path = os.path.join(Path.home(), _("Audiobooks"))
            Storage.create(path=path, default=True)

            os.makedirs(path, exist_ok=True)

        self.add_window(self.ui.window)

        if platform.system().lower() == "linux":
            mpris = MPRIS(self)
            mpris._on_current_changed()

    def handle_exception(self, _):
        print("handle exception")

        exc_type, exc_value, exc_traceback = sys.exc_info()

        if exc_type is SystemExit:
            return

        try:
            reporter.exception("uncaught", exc_value, "\n".join(format_exception(exc_type, exc_value, exc_traceback)))
        finally:
            sys.excepthook(exc_type, exc_value, exc_traceback)

    def quit(self):
        self.app_controller.quit()
        super(Application, self).quit()

    @staticmethod
    def init_custom_widgets():
        FilterListBox()



import re


def collate_natural(s1, s2):
    if s1 == s2:
        return 0

    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    list = sorted([s1, s2], key=alphanum_key)

    if list.index(s1) == 0:
        return -1
    else:
        return 1



import time
import logging
import functools

log = logging.getLogger("timing")

def timing(f):
    @functools.wraps(f)
    def wrap(*args):
        time1 = time.perf_counter()
        ret = f(*args)
        time2 = time.perf_counter()
        log.info('%s function took %.3f ms', f.__name__, (time2-time1)*1000.0)

        return ret
    return wrap


from gettext import gettext as _

from gi.repository import Adw, Gtk

from cozy.ext import inject


HEADER = _("This can have multiple reasons:")
POSSIBILITIES = "\n     • ".join((  # yes, it is a hack, because \t would be too wide
    "",
    _("The audio format is not supported"),
    _("The path or filename contains non utf-8 characters"),
    _("The file(s) are no valid audio files"),
    _("The file(s) are corrupt"),
))

message = HEADER + POSSIBILITIES


class ImportFailedDialog(Adw.MessageDialog):
    """
    Dialog that displays failed files on import.
    """
    main_window = inject.attr("MainWindow")

    def __init__(self, files: list[str]):
        super().__init__(
            heading=_("Some files could not be imported"),
            default_response="cancel",
            close_response="cancel",
            transient_for=self.main_window.window,
            modal=True,
        )

        self.add_response("cancel", _("Ok"))

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=18)
        body_label = Gtk.Label(label=message)

        text_buffer = Gtk.TextBuffer(
            text="\n".join(files).encode("utf-8", errors="replace").decode("utf-8")
        )
        text_view = Gtk.TextView(
            buffer=text_buffer,
            editable=False,
            cursor_visible=False,
            css_classes=["card", "failed-import-card", "monospace"]
        )

        scroller = Gtk.ScrolledWindow(
            max_content_height=200,
            propagate_natural_height=True,
            child=text_view
        )

        box.append(body_label)
        box.append(scroller)
        self.set_extra_child(box)




from gi.repository import Gtk


class DiskElement(Gtk.Box):
    """
    This class represents a small disk number header for the book overview track list.
    """

    def __init__(self, disc_number):
        super().__init__()
        self.add_css_class("dim-label")

        if disc_number > 1:
            self.set_margin_top(18)
        self.set_margin_bottom(3)
        self.set_margin_start(6)

        image = Gtk.Image.new_from_icon_name("media-optical-cd-audio-symbolic")
        self.append(image)

        label = Gtk.Label(margin_start=5)
        text = _("Disc") + " " + str(disc_number)  # TODO: use formatted translation string here
        label.set_markup(f"<b>{text}</b>")
        self.append(label)




from gi.repository import Gtk, Adw

from cozy.ext import inject
from cozy.view_model.app_view_model import AppViewModel
from cozy.view import View

LIBRARY = "main"
EMPTY_STATE = "welcome"
PREPARING_LIBRARY = "import"
BOOK_DETAIL = "book_overview"


class AppView:
    _view_model: AppViewModel = inject.attr(AppViewModel)

    def __init__(self, builder: Gtk.Builder):
        self._builder = builder

        self._get_ui_elements()
        self._connect_view_model()
        self._connect_ui_elements()

        self._update_view_model_view(None, None)

    def _get_ui_elements(self):
        self._main_stack: Gtk.Stack = self._builder.get_object("main_stack")
        self._navigation_view: Adw.NavigationView = self._builder.get_object("navigation_view")

    def _connect_ui_elements(self):
        self._main_stack.connect("notify::visible-child", self._update_view_model_view)
        self._navigation_view.connect("notify::visible-page", self._update_view_model_view)

    def _connect_view_model(self):
        self._view_model.bind_to("view", self._on_view_changed)

    def _on_view_changed(self):
        view = self._view_model.view

        if view == View.EMPTY_STATE:
            self._main_stack.set_visible_child_name(EMPTY_STATE)
        elif view == View.PREPARING_LIBRARY:
            self._main_stack.set_visible_child_name(PREPARING_LIBRARY)
        elif view == View.LIBRARY:
            self._main_stack.set_visible_child_name(LIBRARY)

    def _update_view_model_view(self, *_):
        page = self._main_stack.props.visible_child_name

        if page == LIBRARY:
            if self._navigation_view.props.visible_page.props.tag == BOOK_DETAIL:
                self._view_model.view = View.BOOK_DETAIL
            else:
                self._view_model.view = View.LIBRARY
        elif page == EMPTY_STATE:
            self._view_model.view = View.EMPTY_STATE
        elif page == PREPARING_LIBRARY:
            self._view_model.view = View.PREPARING_LIBRARY




import functools
from typing import Optional

from gi.repository import Adw, Gtk

from cozy.ext import inject
from cozy.ui.widgets.book_element import BookElement
from cozy.ui.delete_book_view import DeleteBookView
from cozy.ui.widgets.filter_list_box import FilterListBox
from cozy.view_model.library_view_model import LibraryViewModel, LibraryViewMode

READER_PAGE = "reader"
AUTHOR_PAGE = "author"
RECENT_PAGE = "recent"
MAIN_BOOK_PAGE = "main"
WELCOME_PAGE = "welcome"

NO_RECENT_PAGE = "no_recent"
BOOKS_PAGE = "books"


class LibraryView:
    _view_model: LibraryViewModel = inject.attr(LibraryViewModel)

    def __init__(self, builder: Gtk.Builder):
        self._builder = builder
        self._connected_book_element: Optional[BookElement] = None

        self._get_ui_elements()
        self._connect_ui_elements()

        self._connect_view_model()

        self.populate_book_box()
        self.populate_author()
        self.populate_reader()

        self._on_library_view_mode_changed()

    def _get_ui_elements(self):
        self._filter_stack: Gtk.Stack = self._builder.get_object("sort_stack")
        self._main_stack: Gtk.Stack = self._builder.get_object("main_stack")
        self._navigation_view: Adw.NavigationView = self._builder.get_object("navigation_view")
        self._split_view: Adw.OverlaySplitView = self._builder.get_object("split_view")
        self._book_box: Gtk.FlowBox = self._builder.get_object("book_box")
        self._filter_stack_revealer: Gtk.Revealer = self._builder.get_object("sort_stack_revealer")
        self._author_box: FilterListBox = self._builder.get_object("author_box")
        self._reader_box: FilterListBox = self._builder.get_object("reader_box")
        self._book_stack: Gtk.Stack = self._builder.get_object("book_stack")

    def _connect_ui_elements(self):
        self._filter_stack.connect("notify::visible-child", self._on_sort_stack_changed)
        self._book_box.set_sort_func(self._view_model.display_book_sort)
        self._book_box.set_filter_func(self._view_model.display_book_filter)

        # We need to connect to row-activated because it will be emitted when the user clicks on a row
        # which is already activated. This is important for the mobile view
        # row-selected is needed because when setting the row using select_row (e.g. when searching), the row-activated
        # signal doesn't emit.
        self._author_box.connect("row-activated", self._on_filter_row_activated)
        self._reader_box.connect("row-activated", self._on_filter_row_activated)
        self._author_box.connect("row-selected", self._on_filter_row_activated)
        self._reader_box.connect("row-selected", self._on_filter_row_activated)

    def _connect_view_model(self):
        self._view_model.bind_to("library_view_mode", self._on_library_view_mode_changed)
        self._view_model.bind_to("authors", self.populate_author)
        self._view_model.bind_to("readers", self.populate_reader)
        self._view_model.bind_to("books", self.populate_book_box)
        self._view_model.bind_to("books-filter", self._book_box.invalidate_filter)
        self._view_model.bind_to("books-filter", self._book_box.invalidate_sort)
        self._view_model.bind_to("selected_filter", self._select_filter_row)
        self._view_model.bind_to("current_book_in_playback", self._current_book_in_playback)
        self._view_model.bind_to("playing", self._playing)
        self._view_model.bind_to("book-progress", self._on_book_progress_changed)

    def _on_sort_stack_changed(self, widget, _):
        page = widget.props.visible_child_name

        view_mode = None
        if page == RECENT_PAGE:
            view_mode = LibraryViewMode.CURRENT
        elif page == AUTHOR_PAGE:
            view_mode = LibraryViewMode.AUTHOR
        elif page == READER_PAGE:
            view_mode = LibraryViewMode.READER

        self._view_model.library_view_mode = view_mode

    def populate_book_box(self):
        self._book_box.remove_all_children()

        for book in self._view_model.books:
            book_element = BookElement(book)
            book_element.connect("play-pause-clicked", self._play_book_clicked)
            book_element.connect("open-book-overview", self._open_book_overview_clicked)
            book_element.connect("book-removed", self._on_book_removed)
            self._book_box.append(book_element)

    def populate_author(self):
        self._author_box.populate(self._view_model.authors)

    def populate_reader(self):
        self._reader_box.populate(self._view_model.readers)

    def _on_library_view_mode_changed(self):
        visible_child_name = None
        active_filter_box: Gtk.ListBox = None

        view_mode = self._view_model.library_view_mode
        main_view_page = MAIN_BOOK_PAGE
        books_view_page = BOOKS_PAGE

        if len(self._view_model.books) < 1:
            main_view_page = WELCOME_PAGE
            visible_child_name = RECENT_PAGE
        elif view_mode == LibraryViewMode.CURRENT:
            visible_child_name = RECENT_PAGE
            if not self._view_model.is_any_book_in_progress:
                books_view_page = NO_RECENT_PAGE
        elif view_mode == LibraryViewMode.AUTHOR:
            visible_child_name = AUTHOR_PAGE
            active_filter_box = self._author_box
        elif view_mode == LibraryViewMode.READER:
            visible_child_name = READER_PAGE
            active_filter_box = self._reader_box

        # https://stackoverflow.com/questions/22178524/gtk-named-stack-childs/22182843#22182843
        self._main_stack.props.visible_child_name = main_view_page
        self._filter_stack.set_visible_child_name(visible_child_name)
        self._book_stack.set_visible_child_name(books_view_page)
        self._navigation_view.pop_to_tag("main")

        if active_filter_box:
            self._apply_selected_filter(active_filter_box.get_selected_row())

        self._invalidate_filters()

    def _invalidate_filters(self):
        self._book_box.invalidate_filter()
        self._book_box.invalidate_sort()

    def _apply_selected_filter(self, row):
        if not row:
            return

        self._view_model.selected_filter = row.data
        self._invalidate_filters()

    def _on_filter_row_activated(self, _, row):
        self._apply_selected_filter(row)

        return True

    def _select_filter_row(self):
        if self._view_model.library_view_mode == LibraryViewMode.AUTHOR:
            self._author_box.select_row_with_content(self._view_model.selected_filter)
        elif self._view_model.library_view_mode == LibraryViewMode.READER:
            self._reader_box.select_row_with_content(self._view_model.selected_filter)

    def _play_book_clicked(self, _, book):
        self._view_model.play_book(book)

    def _open_book_overview_clicked(self, _, book):
        self._view_model.open_book_detail(book)

        return True

    def _on_book_removed(self, _, book):
        if self._view_model.book_files_exist(book):
            DeleteBookView(self._on_book_removed_clicked, book).present()

    def _on_book_removed_clicked(self, _, response, book):
        if response != "delete":
            return

        delete_from_library = True
        delete_files = True  # TODO: maybe an option to not delete the files

        if delete_files:
            self._view_model.delete_book_files(book)

        if delete_from_library:
            self._view_model.remove_book(book)

    def _current_book_in_playback(self):
        if self._connected_book_element:
            self._connected_book_element.set_playing(False)

        self._connected_book_element = None

        index = 0
        while book_element := self._book_box.get_child_at_index(index):
            if book_element.book == self._view_model.current_book_in_playback:
                self._connected_book_element = book_element
                break
            index += 1

    def _playing(self):
        if self._connected_book_element:
            self._connected_book_element.set_playing(self._view_model.playing)

    def _on_book_progress_changed(self):
        if not self._connected_book_element:
            return

        self._connected_book_element.update_progress()



from gi.repository import Gtk, GObject, Gdk, Gio

from cozy.model.book import Book
from cozy.ui.widgets.album_element import AlbumElement


@Gtk.Template.from_resource('/com/github/geigi/cozy/book_element.ui')
class BookElement(Gtk.FlowBoxChild):
    __gtype_name__ = "BookElement"

    name_label: Gtk.Label = Gtk.Template.Child()
    author_label: Gtk.Label = Gtk.Template.Child()
    container_box: Gtk.Box = Gtk.Template.Child()

    def __init__(self, book: Book):
        super().__init__()

        self.book = book
        self.pressed = False

        self.name_label.set_text(book.name)
        self.name_label.set_tooltip_text(book.name)
        self.author_label.set_text(book.author)
        self.author_label.set_tooltip_text(book.author)

        self.art = AlbumElement(self.book)
        self.art.connect("play-pause-clicked", self._on_album_art_press_event)

        self.container_box.prepend(self.art)
        self.set_cursor(Gdk.Cursor.new_from_name("pointer"))

        self._add_event_controllers()

    def _add_event_controllers(self):
        primary_button_gesture = Gtk.GestureClick(button=Gdk.BUTTON_PRIMARY)
        # primary_button_gesture.connect("pressed", self._select_item)
        primary_button_gesture.connect("released", self._open_book_overview)
        self.container_box.add_controller(primary_button_gesture)

        secondary_button_gesture = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        secondary_button_gesture.connect("released", self._show_context_menu)
        self.container_box.add_controller(secondary_button_gesture)

        # FIXME: When clicking on an album's play button in the recents view,
        # it jumps to the first position, and GtkGestureLongPress thinks it's
        # a long press gesture, although it's just an unfinished long press
        long_press_gesture = Gtk.GestureLongPress()
        long_press_gesture.connect("pressed", self._show_context_menu)
        self.container_box.add_controller(long_press_gesture)

        key_event_controller = Gtk.EventControllerKey()
        key_event_controller.connect("key-pressed", self._on_key_press_event)
        self.container_box.add_controller(key_event_controller)

        motion_event_controller = Gtk.EventControllerMotion()
        motion_event_controller.connect("enter", self._on_cover_enter_notify)
        motion_event_controller.connect("leave", self._on_cover_leave_notify)
        self.container_box.add_controller(motion_event_controller)

    def set_playing(self, is_playing):
        self.art.set_playing(is_playing)

    def update_progress(self):
        self.art.update_progress()

    def _create_context_menu(self):
        menu_model = Gio.Menu()

        self.install_action("book_element.mark_as_read", None, self._mark_as_read)
        menu_model.append(_("Mark as read"), "book_element.mark_as_read")

        self.install_action("book_element.jump_to_folder", None, self._jump_to_folder)
        menu_model.append(_("Open in file browser"), "book_element.jump_to_folder")

        self.install_action("book_element.remove_book", None, self._remove_book)
        menu_model.append(_("Remove from library"), "book_element.remove_book")

        menu = Gtk.PopoverMenu(menu_model=menu_model, has_arrow=False)
        menu.set_parent(self.art)

        return menu

    def _remove_book(self, *_):
        self.emit("book-removed", self.book)

    def _mark_as_read(self, *_):
        self.book.position = -1

    def _jump_to_folder(self, *_):
        """
        Opens the folder containing this books files in the default file explorer.
        """
        track = self.book.chapters[0]

        file_launcher = Gtk.FileLauncher(file=Gio.File.new_for_path(track.file))
        dummy_callback = lambda d, r: d.open_containing_folder_finish(r)
        file_launcher.open_containing_folder(None, None, dummy_callback)

    def _show_context_menu(self, gesture: Gtk.Gesture, *_):
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)

        self._create_context_menu().popup()

    def _select_item(self, gesture: Gtk.Gesture, *_):
        if super().get_sensitive():
            gesture.set_state(Gtk.EventSequenceState.CLAIMED)
            self.pressed = True
            self.container_box.add_css_class("selected")

    def _open_book_overview(self, gesture, *_):
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)

        self.pressed = False
        self.container_box.remove_css_class("selected")
        if super().get_sensitive():
            self.emit("open-book-overview", self.book)

    def _on_key_press_event(self, keyval, *_):
        if keyval == Gdk.KEY_Return and super().get_sensitive():
            self.emit("open-book-overview", self.book)

    def _on_cover_enter_notify(self, *_):
        self.art.set_hover(True)

    def _on_cover_leave_notify(self, *_):
        self.art.set_hover(False)

    def _on_album_art_press_event(self, *_):
        self.emit("play-pause-clicked", self.book)


GObject.type_register(BookElement)
GObject.signal_new('play-pause-clicked', BookElement, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT,
                   (GObject.TYPE_PYOBJECT,))
GObject.signal_new('open-book-overview', BookElement, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT,
                   (GObject.TYPE_PYOBJECT,))
GObject.signal_new('book-removed', BookElement, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT,
                   (GObject.TYPE_PYOBJECT,))




from gi.repository import Gtk


def remove_all_children(self):
    """
    Removes all widgets from a gtk widget.
    """
    self.set_visible(False)

    child = self.get_first_child()
    while child:
        next = child.get_next_sibling()
        self.remove(child)
        child = next

    self.set_visible(True)


def extend_gtk_container():
    Gtk.Widget.remove_all_children = remove_all_children



import threading
from typing import Callable, Sequence

from gi.repository import Adw, Gtk

from cozy.ext import inject
from cozy.model.book import Book
from cozy.ui.headerbar import Headerbar
from cozy.ui.widgets.book_row import BookRow
from cozy.ui.widgets.search_results import ArtistResultRow
from cozy.view_model.search_view_model import SearchViewModel


@Gtk.Template.from_resource("/com/github/geigi/cozy/search_page.ui")
class SearchView(Adw.Bin):
    __gtype_name__ = "SearchView"

    stack: Gtk.Stack = Gtk.Template.Child()
    search_scroller: Gtk.ScrolledWindow = Gtk.Template.Child()
    start_searching_page: Adw.StatusPage = Gtk.Template.Child()
    nothing_found_page: Adw.StatusPage = Gtk.Template.Child()

    book_result_box: Adw.PreferencesGroup = Gtk.Template.Child()
    author_result_box: Adw.PreferencesGroup = Gtk.Template.Child()
    reader_result_box: Adw.PreferencesGroup = Gtk.Template.Child()

    book_result_list: Gtk.ListBox = Gtk.Template.Child()
    author_result_list: Gtk.ListBox = Gtk.Template.Child()
    reader_result_list: Gtk.ListBox = Gtk.Template.Child()

    view_model = inject.attr(SearchViewModel)
    main_view = inject.attr("MainWindow")

    search_thread: threading.Thread

    def __init__(self, main_window_builder: Gtk.Builder, headerbar: Headerbar) -> None:
        super().__init__()

        self.library_stack: Gtk.Stack = main_window_builder.get_object("library_stack")
        self.library_stack.add_child(self)

        self.split_view: Gtk.Stack = main_window_builder.get_object("split_view")

        self.search_bar = headerbar.search_bar
        self.entry = headerbar.search_entry
        self.entry.connect("search-changed", self._on_search_changed)

        self.search_thread = threading.Thread(target=self.view_model.search)

        self.view_model.bind_to("close", self.close)
        self.main_view.create_action("search", self.open, ["<primary>f"])

    def open(self, *_) -> None:
        self.library_stack.set_visible_child(self)
        self.search_bar.set_search_mode(True)
        self.main_view.play_pause_action.set_enabled(False)

    def close(self) -> None:
        self.library_stack.set_visible_child(self.split_view)
        self.search_bar.set_search_mode(False)
        self.main_view.play_pause_action.set_enabled(True)

    def on_state_changed(self, widget: Gtk.Widget, param) -> None:
        if widget.get_property(param.name):
            self.open()
        else:
            self.close()

    def _on_search_changed(self, _) -> None:
        search_query = self.entry.get_text()
        if not search_query:
            self.stack.set_visible_child(self.start_searching_page)
            return

        if self.search_thread.is_alive():
            self.search_thread.join(timeout=0.1)

        self.search_thread = threading.Thread(
            target=self.view_model.search, args=(search_query, self._display_results)
        )
        self.search_thread.start()

    def _display_results(self, books: list[Book], authors: list[str], readers: list[str]) -> None:
        if not any((books, authors, readers)):
            self.stack.set_visible_child(self.nothing_found_page)
            return

        self.stack.set_visible_child(self.search_scroller)
        self._populate_listbox(
            books, self.book_result_list, self.book_result_box, self.view_model.jump_to_book
        )
        self._populate_listbox(
            authors, self.author_result_list, self.author_result_box, self.view_model.jump_to_author
        )
        self._populate_listbox(
            readers, self.reader_result_list, self.reader_result_box, self.view_model.jump_to_reader
        )

    def _populate_listbox(
        self,
        results: Sequence[str | Book],
        listbox: Gtk.ListBox,
        box: Adw.PreferencesGroup,
        callback: Callable[[str | Book], None],
    ) -> None:
        box.set_visible(False)
        listbox.remove_all()

        if not results:
            return

        if isinstance(results[0], Book):
            row_type = BookRow
        else:
            row_type = ArtistResultRow

        for result in results:
            listbox.append(row_type(result, callback))

        box.set_visible(True)



from typing import Callable

from gi.repository import GLib

import cozy.ext.inject as inject
from cozy.application_settings import ApplicationSettings
from cozy.architecture.event_sender import EventSender
from cozy.architecture.observable import Observable
from cozy.control.filesystem_monitor import FilesystemMonitor
from cozy.extensions.set import split_strings_to_set
from cozy.model.book import Book
from cozy.model.library import Library
from cozy.open_view import OpenView


class SearchViewModel(Observable, EventSender):
    _fs_monitor: FilesystemMonitor = inject.attr("FilesystemMonitor")
    _model: Library = inject.attr(Library)
    _application_settings: ApplicationSettings = inject.attr(ApplicationSettings)

    def __init__(self):
        super().__init__()
        super(Observable, self).__init__()

    def _get_available_books(self) -> list[Book]:
        is_book_online = self._fs_monitor.get_book_online

        if self._application_settings.hide_offline:
            return [book for book in self._model.books if is_book_online(book)]
        else:
            return self._model.books

    def search(
        self, search_query: str, callback: Callable[[list[Book], list[str], list[str]], None]
    ) -> None:
        search_query = search_query.lower()

        available_books = self._get_available_books()
        books = {
            book
            for book in available_books
            if search_query in book.name.lower()
            or search_query in book.author.lower()
            or search_query in book.reader.lower()
        }

        available_book_authors = split_strings_to_set({book.author for book in available_books})
        authors = {author for author in available_book_authors if search_query in author.lower()}

        available_book_readers = split_strings_to_set({book.reader for book in available_books})
        readers = {reader for reader in available_book_readers if search_query in reader.lower()}

        GLib.MainContext.default().invoke_full(
            GLib.PRIORITY_DEFAULT,
            callback,
            sorted(books, key=lambda book: book.name.lower()),
            sorted(authors),
            sorted(readers),
        )

    def close(self) -> None:
        self._notify("close")

    def jump_to_book(self, book: Book) -> None:
        self.emit_event(OpenView.BOOK, book)
        self.close()

    def jump_to_author(self, author: str) -> None:
        self.emit_event(OpenView.AUTHOR, author)
        self.close()

    def jump_to_reader(self, reader: str) -> None:
        self.emit_event(OpenView.READER, reader)
        self.close()



from typing import List, Callable

import gi
from gi.repository import GLib


class EventSender:
    _listeners: List[Callable]

    def __init__(self):
        self._listeners = []

    def emit_event(self, event, message=None):
        if isinstance(event, tuple) and message is None:
            event, message = event

        for function in self._listeners:
            function(event, message)

    def emit_event_main_thread(self, event: str, message=None):
        GLib.MainContext.default().invoke_full(GLib.PRIORITY_DEFAULT_IDLE, self.emit_event, (event, message))

    def add_listener(self, function: Callable[[str, object], None]):
        self._listeners.append(function)

    def destroy_listeners(self):
        self._listeners = []



import logging

import gi

from cozy.control.artwork_cache import ArtworkCache
from cozy.db.book import Book
from cozy.ext import inject
from cozy.ui.widgets.playback_speed_popover import PlaybackSpeedPopover
from cozy.ui.widgets.seek_bar import SeekBar
from cozy.ui.widgets.sleep_timer import SleepTimer
from cozy.view_model.playback_control_view_model import PlaybackControlViewModel

from gi.repository import Adw, Gtk, Gdk

log = logging.getLogger("MediaController")

COVER_SIZE = 46


@Gtk.Template.from_resource('/com/github/geigi/cozy/media_controller.ui')
class MediaController(Adw.BreakpointBin):
    __gtype_name__ = "MediaController"

    seek_bar_container: Gtk.Box = Gtk.Template.Child()

    play_button: Gtk.Button = Gtk.Template.Child()
    prev_button: Gtk.Button = Gtk.Template.Child()
    next_button: Gtk.Button = Gtk.Template.Child()
    volume_button: Gtk.ScaleButton = Gtk.Template.Child()

    cover_img: Gtk.Image = Gtk.Template.Child()
    title_label: Gtk.Label = Gtk.Template.Child()
    subtitle_label: Gtk.Label = Gtk.Template.Child()

    playback_speed_button: Gtk.MenuButton = Gtk.Template.Child()
    timer_button: Gtk.MenuButton = Gtk.Template.Child()

    timer_image: Gtk.Image = Gtk.Template.Child()

    def __init__(self, main_window_builder: Gtk.Builder):
        super().__init__()

        self.container_bar: Gtk.Revealer = main_window_builder.get_object("media_control_box")
        self.container_bar.set_child(self)

        self.seek_bar = SeekBar()
        self.seek_bar_container.append(self.seek_bar)

        self.sleep_timer: SleepTimer = SleepTimer(self.timer_image)
        self.playback_speed_button.set_popover(PlaybackSpeedPopover())
        self.timer_button.set_popover(self.sleep_timer)

        self._playback_control_view_model: PlaybackControlViewModel = inject.instance(PlaybackControlViewModel)
        self._artwork_cache: ArtworkCache = inject.instance(ArtworkCache)
        self._connect_view_model()
        self._connect_widgets()

        self._on_book_changed()
        self._on_lock_ui_changed()
        self._on_length_changed()
        self._on_position_changed()
        self._on_volume_changed()

    def _connect_view_model(self):
        self._playback_control_view_model.bind_to("book", self._on_book_changed)
        self._playback_control_view_model.bind_to("playing", self._on_play_changed)
        self._playback_control_view_model.bind_to("length", self._on_length_changed)
        self._playback_control_view_model.bind_to("position", self._on_position_changed)
        self._playback_control_view_model.bind_to("lock_ui", self._on_lock_ui_changed)
        self._playback_control_view_model.bind_to("volume", self._on_volume_changed)

    def _connect_widgets(self):
        self.play_button.connect("clicked", self._play_clicked)
        self.volume_button.connect("value-changed", self._on_volume_button_changed)
        self.seek_bar.connect("position-changed", self._on_seek_bar_position_changed)

        self.prev_button.connect("clicked", self._rewind_clicked)
        self.next_button.connect("clicked", self._forward_clicked)
        self.seek_bar.connect("rewind", self._rewind_clicked)
        self.seek_bar.connect("forward", self._forward_clicked)

        cover_click_gesture = Gtk.GestureClick()
        cover_click_gesture.connect("pressed", self._cover_clicked)
        self.cover_img.add_controller(cover_click_gesture)
        self.cover_img.set_cursor(Gdk.Cursor.new_from_name("pointer"))

    def _set_cover_image(self, book: Book):
        paintable = self._artwork_cache.get_cover_paintable(book, self.get_scale_factor(), COVER_SIZE)
        if paintable:
            self.cover_img.set_from_paintable(paintable)
        else:
            self.cover_img.set_from_icon_name("book-open-variant-symbolic")
            self.cover_img.props.pixel_size = COVER_SIZE

    def _on_book_changed(self) -> None:
        book = self._playback_control_view_model.book
        self._set_book(book)
        self._show_media_information(bool(book))

    def _show_media_information(self, visibility: bool) -> None:
        self.title_label.set_visible(visibility)
        self.subtitle_label.set_visible(visibility)
        self.cover_img.set_visible(visibility)
        self.seek_bar.visible = visibility

    def _set_book(self, book: Book) -> None:
        if book is not None:
            self._set_cover_image(book)
            self.title_label.set_text(book.name)
            self.title_label.set_tooltip_text(book.name)
            self.subtitle_label.set_text(book.current_chapter.name)
            self.subtitle_label.set_tooltip_text(book.current_chapter.name)

    def _on_play_changed(self):
        if self._playback_control_view_model.playing:
            self.play_button.set_icon_name("media-playback-pause-symbolic")
        else:
            self.play_button.set_icon_name("media-playback-start-symbolic")

    def _on_position_changed(self):
        position = self._playback_control_view_model.relative_position
        if position is not None:
            self.seek_bar.position = position

    def _on_length_changed(self):
        length = self._playback_control_view_model.length
        if length:
            self.seek_bar.length = length

    def _on_lock_ui_changed(self):
        sensitive = not self._playback_control_view_model.lock_ui
        self.container_bar.set_reveal_child(sensitive)

    def _on_volume_changed(self):
        self.volume_button.set_value(self._playback_control_view_model.volume)

    def _play_clicked(self, *_):
        self._playback_control_view_model.play_pause()

    def _rewind_clicked(self, *_):
        self._playback_control_view_model.rewind()

    def _forward_clicked(self, *_):
        self._playback_control_view_model.forward()

    def _cover_clicked(self, *_):
        self._playback_control_view_model.open_book_detail()

    def _on_volume_button_changed(self, _, volume):
        self._playback_control_view_model.volume = volume

    def _on_seek_bar_position_changed(self, _, position):
        self._playback_control_view_model.relative_position = position



from gi.repository import Gtk, Pango, GObject, Gdk

from cozy.control.string_representation import seconds_to_str
from cozy.model.chapter import Chapter


@Gtk.Template.from_resource('/com/github/geigi/cozy/chapter_element.ui')
class ChapterElement(Gtk.Box):
    __gtype_name__ = "ChapterElement"

    icon_stack: Gtk.Stack = Gtk.Template.Child()
    number_label: Gtk.Label = Gtk.Template.Child()
    play_icon: Gtk.Image = Gtk.Template.Child()
    title_label: Gtk.Label = Gtk.Template.Child()
    duration_label: Gtk.Label = Gtk.Template.Child()

    def __init__(self, chapter: Chapter):
        self.selected = False
        self.chapter = chapter

        super().__init__()

        gesture = Gtk.GestureClick(button=Gdk.BUTTON_PRIMARY)
        gesture.connect("released", self._on_button_press)
        self.add_controller(gesture)

        motion = Gtk.EventControllerMotion()
        motion.connect("enter", self._on_enter_notify)
        motion.connect("leave", self._on_leave_notify)
        self.add_controller(motion)

        self.number_label.set_text(str(self.chapter.number))
        self.title_label.set_text(self.chapter.name)
        self.duration_label.set_text(seconds_to_str(self.chapter.length))

    def _on_button_press(self, *_):
        self.emit("play-pause-clicked", self.chapter)

    def _on_enter_notify(self, *_):
        self.add_css_class("box_hover")
        self.play_icon.add_css_class("box_hover")

        if not self.selected:
            self.icon_stack.set_visible_child_name("play_icon")

    def _on_leave_notify(self, *_):
        self.remove_css_class("box_hover")
        self.play_icon.remove_css_class("box_hover")

        if not self.selected:
            self.icon_stack.set_visible_child_name("number")

    def select(self):
        self.selected = True
        self.icon_stack.set_visible_child_name("play_icon")

    def deselect(self):
        self.selected = False
        self.icon_stack.set_visible_child_name("number")

    def set_playing(self, playing):
        if playing:
            self.play_icon.set_from_icon_name("media-playback-pause-symbolic")
        else:
            self.play_icon.set_from_icon_name("media-playback-start-symbolic")


GObject.type_register(ChapterElement)
GObject.signal_new('play-pause-clicked', ChapterElement, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT,
                   (GObject.TYPE_PYOBJECT,))



from gi.repository import Adw, Gtk

from cozy.ext import inject


class ToastNotifier:
    _builder: Gtk.Builder = inject.attr("MainWindowBuilder")

    def __init__(self) -> None:
        super().__init__()

        self.overlay: Adw.ToastOverlay = self._builder.get_object("toast_overlay")

    def show(self, message: str) -> None:
        self.overlay.add_toast(Adw.Toast(title=message, timeout=2))




from gi.repository import Adw, Gtk, GObject

from .storages import ask_storage_location

from typing import Callable


@Gtk.Template.from_resource('/com/github/geigi/cozy/first_import_button.ui')
class FirstImportButton(Gtk.Button):
    __gtype_name__ = "FirstImportButton"

    stack: Gtk.Stack = Gtk.Template.Child()
    label: Adw.ButtonContent = Gtk.Template.Child()
    spinner: Gtk.Spinner = Gtk.Template.Child()

    def __init__(self, callback: Callable[[str], None], initial_folder: str) -> None:
        super().__init__()

        self.connect("clicked", lambda *_: ask_storage_location(callback, initial_folder))

    def disable(self) -> None:
        self.set_sensitive(False)
        self.spinner.set_spinning(True)
        self.stack.set_visible_child(self.spinner)



import logging

from gi.repository import Adw, Gtk

from cozy.application_settings import ApplicationSettings
from cozy.architecture.event_sender import EventSender
from cozy.architecture.observable import Observable
from cozy.ext import inject
from cozy.media.importer import Importer
from cozy.model.settings import Settings

log = logging.getLogger("settings_view_model")


class SettingsViewModel(Observable, EventSender):
    _importer: Importer = inject.attr(Importer)
    _model: Settings = inject.attr(Settings)
    _app_settings: ApplicationSettings = inject.attr(ApplicationSettings)

    def __init__(self):
        super().__init__()
        super(Observable, self).__init__()

        self._lock_ui: bool = False

        self.style_manager = Adw.StyleManager.get_default()
        self._set_dark_mode()

        self._app_settings.add_listener(self._on_app_setting_changed)

        if self._model.first_start:
            self._importer.scan()

    @property
    def lock_ui(self) -> bool:
        return self._lock_ui

    @lock_ui.setter
    def lock_ui(self, new_value: bool):
        self._lock_ui = new_value
        self._notify("lock_ui")

    def _set_dark_mode(self):
        if self._app_settings.dark_mode:
            self.style_manager.set_color_scheme(Adw.ColorScheme.PREFER_DARK)
        else:
            self.style_manager.set_color_scheme(Adw.ColorScheme.PREFER_LIGHT)

    def _on_app_setting_changed(self, event: str, data):
        if event == "dark-mode":
            self._set_dark_mode()



import logging

import gi

from cozy.ext import inject
from cozy.ui.widgets.progress_popover import ProgressPopover
from cozy.view_model.headerbar_view_model import HeaderbarViewModel, HeaderBarState

from gi.repository import Adw, Gtk, GObject

from cozy.ext import inject
from cozy.ui.widgets.progress_popover import ProgressPopover
from cozy.view_model.headerbar_view_model import HeaderBarState, HeaderbarViewModel

log = logging.getLogger("Headerbar")


@Gtk.Template.from_resource("/com/github/geigi/cozy/headerbar.ui")
class Headerbar(Gtk.Box):
    __gtype_name__ = "Headerbar"

    headerbar: Adw.HeaderBar = Gtk.Template.Child()

    search_bar: Gtk.SearchBar = Gtk.Template.Child()
    search_entry: Gtk.SearchEntry = Gtk.Template.Child()

    show_sidebar_button: Gtk.ToggleButton = Gtk.Template.Child()
    search_button: Gtk.MenuButton = Gtk.Template.Child()
    menu_button: Gtk.MenuButton = Gtk.Template.Child()

    progress_menu_button: Gtk.MenuButton = Gtk.Template.Child()
    progress_spinner: Gtk.Spinner = Gtk.Template.Child()

    view_switcher: Adw.ViewSwitcher = Gtk.Template.Child()

    def __init__(self, main_window_builder: Gtk.Builder):
        super().__init__()

        self.header_container: Adw.ToolbarView = main_window_builder.get_object("header_container")
        self.header_container.add_top_bar(self)

        self.mobile_view_switcher: Adw.ViewSwitcherBar = main_window_builder.get_object(
            "mobile_view_switcher"
        )
        self.split_view: Adw.OverlaySplitView = main_window_builder.get_object("split_view")

        self.sort_stack: Adw.ViewStack = main_window_builder.get_object("sort_stack")
        self.view_switcher.set_stack(self.sort_stack)
        self.mobile_view_switcher.set_stack(self.sort_stack)

        self.progress_popover = ProgressPopover()
        self.progress_menu_button.set_popover(self.progress_popover)

        self.search_bar.connect_entry(self.search_entry)
        self.search_bar.set_key_capture_widget(self.header_container)

        self._headerbar_view_model: HeaderbarViewModel = inject.instance(HeaderbarViewModel)
        self._connect_view_model()
        self._connect_widgets()

        self._set_show_sidebar_button_visible()

    def _connect_view_model(self):
        self._headerbar_view_model.bind_to("state", self._on_state_changed)
        self._headerbar_view_model.bind_to("work_progress", self._on_work_progress_changed)
        self._headerbar_view_model.bind_to("work_message", self._on_work_message_changed)
        self._headerbar_view_model.bind_to("lock_ui", self._on_lock_ui_changed)

    def _connect_widgets(self):
        self.split_view.connect("notify::show-sidebar", self._on_sidebar_toggle)
        self.show_sidebar_button.connect("notify::active", self._on_sidebar_toggle)
        self.mobile_view_switcher.connect("notify::reveal", self._on_mobile_view)
        self.sort_stack.connect("notify::visible-child", self._set_show_sidebar_button_visible)

        self.mobile_view_switcher.bind_property(
            "reveal", self.split_view, "collapsed", GObject.BindingFlags.SYNC_CREATE
        )

    def _on_mobile(self) -> bool:
        return self.mobile_view_switcher.props.reveal

    def _set_show_sidebar_button_visible(self, *_):
        page = self.sort_stack.props.visible_child_name

        if not self._on_mobile():
            self.show_sidebar_button.set_visible(False)
            self.split_view.set_collapsed(False)
            self.split_view.set_show_sidebar(page != "recent")
            return

        if self.mobile_view_switcher.props.reveal:
            self.show_sidebar_button.set_visible(page != "recent")
        else:
            self.show_sidebar_button.set_visible(False)

        self.search_button.set_active(False)

    def _on_mobile_view(self, widget, _):
        page = self.sort_stack.props.visible_child_name

        if widget.props.reveal:
            self.headerbar.set_title_widget(Adw.WindowTitle(title="Cozy"))
        else:
            self.headerbar.set_title_widget(self.view_switcher)

        self._set_show_sidebar_button_visible()

    def _on_sidebar_toggle(self, widget, param):
        show_sidebar = widget.get_property(param.name)

        if widget is self.show_sidebar_button:
            self.split_view.set_show_sidebar(show_sidebar)
        elif widget is self.split_view:
            self.show_sidebar_button.set_active(show_sidebar)

    def _on_state_changed(self):
        if self._headerbar_view_model.state == HeaderBarState.PLAYING:
            self.progress_menu_button.set_visible(False)
            self.progress_popover.set_progress(0)
            self.progress_spinner.stop()
        else:
            self.progress_menu_button.set_visible(True)
            self.progress_spinner.start()

    def _on_work_progress_changed(self):
        self.progress_popover.set_progress(self._headerbar_view_model.work_progress)

    def _on_work_message_changed(self):
        self.progress_popover.set_message(self._headerbar_view_model.work_message)

    def _on_lock_ui_changed(self):
        self.search_button.set_sensitive(not self._headerbar_view_model.lock_ui)



import webbrowser

import gi
from gi.repository import Adw

EXPLANATION = _("During an update of the database an error occurred and Cozy will not be able to startup.\
 A backup of the database was created before the update and has been restored now.\
 Until this issue is resolved please use version 0.9.5 of Cozy.\
 You can help resolve this problem by reporting an issue on GitHub.")


class DBMigrationFailedView(Adw.MessageDialog):
    def __init__(self):
        super().__init__(
            heading=_("Failed to Update Database"),
            body=EXPLANATION,
            default_response="help",
            close_response="close",
            modal=True,
        )

        self.add_response("close", _("Close Cozy"))
        self.add_response("help", _("Receive help on GitHub"))
        self.set_response_appearance("help", Adw.ResponseAppearance.SUGGESTED)

        self.connect("response", self.get_help)

    def get_help(self, *_, response):
        if response == "help":
            webbrowser.open("https://github.com/geigi/cozy/issues", new=2)




# Forked from https://github.com/gnumdk/lollypop/blob/master/lollypop/mpris.py
# copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# copyright (c) 2016 Gaurav Narula
# copyright (c) 2016 Felipe Borges <felipeborges@gnome.org>
# copyright (c) 2013 Arnel A. Borja <kyoushuu@yahoo.com>
# copyright (c) 2013 Vadim Rutkovsky <vrutkovs@redhat.com>
# copyright (c) 2017 Julian Geywitz <cozy@geigi.de>
# copyright (c) 2023 Benedek Dévényi <rdbende@proton.me>

import logging
import re
import time
from dataclasses import dataclass

from gi.repository import Gio, GLib

from cozy.application_settings import ApplicationSettings
from cozy.control.artwork_cache import ArtworkCache
from cozy.ext import inject
from cozy.media.player import NS_TO_SEC, US_TO_SEC, Player
from cozy.model.book import Book
from cozy.report import reporter

log = logging.getLogger("mpris")

CamelCasePattern = re.compile(r"(?<!^)(?=[A-Z])")


def to_snake_case(name: str) -> str:
    return CamelCasePattern.sub("_", name).lower()


@dataclass(kw_only=True, frozen=True, slots=True)
class Metadata:
    track_id: str
    track_number: int
    title: str
    album: str
    artist: list[str]
    length: int
    url: str
    artwork_uri: str

    def to_dict(self) -> dict[str, GLib.Variant]:
        data = {}
        data["mpris:trackid"] = GLib.Variant("o", self.track_id)
        data["xesam:trackNumber"] = GLib.Variant("i", self.track_number)
        data["xesam:title"] = GLib.Variant("s", self.title)
        data["xesam:album"] = GLib.Variant("s", self.album)
        data["xesam:artist"] = GLib.Variant("as", self.artist)
        data["mpris:length"] = GLib.Variant("x", self.length)
        data["xesam:url"] = GLib.Variant("s", self.url)

        if self.artwork_uri:
            data["mpris:artUrl"] = GLib.Variant("s", "file://" + self.artwork_uri)

        return data

    @staticmethod
    def no_track() -> dict[str, GLib.Variant]:
        no_track_path = GLib.Variant("o", "/org/mpris/MediaPlayer2/TrackList/NoTrack")
        return {"mpris:trackid": no_track_path}


class Server:
    def __init__(self, connection: Gio.DBusConnection, path: str) -> None:
        self.method_outargs = {}
        self.method_inargs = {}

        for interface in Gio.DBusNodeInfo.new_for_xml(self.__doc__).interfaces:
            for method in interface.methods:
                self.method_inargs[method.name] = tuple(
                    arg.signature for arg in method.in_args
                )
                out_sig = [arg.signature for arg in method.out_args]
                self.method_outargs[method.name] = "(" + "".join(out_sig) + ")"

            try:
                connection.register_object(
                    object_path=path,
                    interface_info=interface,
                    method_call_closure=self.on_method_call,
                )
            except Exception:
                log.error("MPRIS is already connected from another Cozy process.")

    def on_method_call(
        self,
        connection: Gio.DBusConnection,
        sender: str,
        object_path: str,
        interface_name: str,
        method_name: str,
        parameters: GLib.Variant,
        invocation: Gio.DBusMethodInvocation,
    ) -> None:
        args = list(parameters.unpack())
        for i, sig in enumerate(self.method_inargs[method_name]):
            if sig == "h":
                msg = invocation.get_message()
                fd_list = msg.get_unix_fd_list()
                args[i] = fd_list.get(args[i])

        snake_method = to_snake_case(method_name)
        try:
            result = getattr(self, snake_method)(*args)
        except AttributeError:
            invocation.return_dbus_error(
                f"{interface_name}.Error.NotSupported", "Unsupported property"
            )
        except Exception as e:
            log.error(e)
            reporter.exception("mpris", e)
            reporter.error(
                "mpris",
                f"MPRIS method call failed with method name: {method_name}",
            )
            invocation.return_dbus_error(
                f"{interface_name}.Error.Failed", "Internal exception occurred"
            )
        else:
            # out_args is at least (signature1).
            # We therefore always wrap the result as a tuple.
            # Reference:
            # https://bugzilla.gnome.org/show_bug.cgi?id=765603
            result = (result,)

            out_args = self.method_outargs[method_name]
            if out_args != "()" and result[0]:
                variant = GLib.Variant(out_args, result)
                invocation.return_value(variant)
            else:
                invocation.return_value(None)


class MPRIS(Server):
    """
    <node xmlns:doc="http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
      <interface name="org.freedesktop.DBus.Introspectable">
        <method name="Introspect">
          <arg name="data" direction="out" type="s"/>
        </method>
      </interface>
      <interface name="org.freedesktop.DBus.Properties">
        <method name="Get">
          <arg name="interface" direction="in" type="s"/>
          <arg name="property" direction="in" type="s"/>
          <arg name="value" direction="out" type="v"/>
        </method>
        <method name="Set">
          <arg name="interface_name" direction="in" type="s"/>
          <arg name="property_name" direction="in" type="s"/>
          <arg name="value" direction="in" type="v"/>
        </method>
        <method name="GetAll">
          <arg name="interface" direction="in" type="s"/>
          <arg name="properties" direction="out" type="a{sv}"/>
        </method>
        <signal name="PropertiesChanged">
          <arg name="interface_name" type="s" />
          <arg name="changed_properties" type="a{sv}" />
          <arg name="invalidated_properties" type="as" />
        </signal>
      </interface>
      <interface name="org.mpris.MediaPlayer2">
        <method name="Raise"/>
        <method name="Quit"/>
        <property name="CanQuit" type="b" access="read" />
        <property name="Fullscreen" type="b" access="readwrite" />
        <property name="CanRaise" type="b" access="read" />
        <property name="HasTrackList" type="b" access="read"/>
        <property name="Identity" type="s" access="read"/>
        <property name="DesktopEntry" type="s" access="read"/>
        <property name="SupportedUriSchemes" type="as" access="read"/>
        <property name="SupportedMimeTypes" type="as" access="read"/>
      </interface>
      <interface name="org.mpris.MediaPlayer2.Player">
        <method name="Next"/>
        <method name="Previous"/>
        <method name="Pause"/>
        <method name="PlayPause"/>
        <method name="Stop"/>
        <method name="Play"/>
        <method name="Seek">
          <arg direction="in" name="Offset" type="x"/>
        </method>
        <method name="SetPosition">
          <arg direction="in" name="TrackId" type="o"/>
          <arg direction="in" name="Position" type="x"/>
        </method>
        <method name="OpenUri">
          <arg direction="in" name="Uri" type="s"/>
        </method>
        <signal name="Seeked">
          <arg name="Position" type="x"/>
        </signal>
        <property name="PlaybackStatus" type="s" access="read"/>
        <property name="Metadata" type="a{sv}" access="read"/>
        <property name="Position" type="x" access="read"/>
        <property name="CanGoNext" type="b" access="read"/>
        <property name="CanGoPrevious" type="b" access="read"/>
        <property name="CanPlay" type="b" access="read"/>
        <property name="CanPause" type="b" access="read"/>
        <property name="CanSeek" type="b" access="read"/>
        <property name="CanControl" type="b" access="read"/>
        <property name="Volume" type="d" access="readwrite"/>
      </interface>
    </node>
    """

    MEDIA_PLAYER2_INTERFACE = "org.mpris.MediaPlayer2"
    MEDIA_PLAYER2_PLAYER_INTERFACE = "org.mpris.MediaPlayer2.Player"

    _player: Player = inject.attr(Player)
    _artwork_cache: ArtworkCache = inject.attr(ArtworkCache)
    _app_settings: ApplicationSettings = inject.attr(ApplicationSettings)

    def __init__(self, app) -> None:
        self._bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        Gio.bus_own_name_on_connection(
            self._bus,
            "org.mpris.MediaPlayer2.Cozy",
            Gio.BusNameOwnerFlags.NONE,
            None,
            None,
        )
        super().__init__(self._bus, "/org/mpris/MediaPlayer2")

        self._application = app
        self._metadata = self._get_new_metadata()

        self._player.add_listener(self._on_player_changed)
        self._app_settings.add_listener(self._on_app_setting_changed)

    def introspect(self):
        return self.__doc__

    def quit(self):
        self._application.quit()

    def next(self):
        self._player.forward()

    def previous(self):
        self._player.rewind()

    def play(self):
        self._player.play_pause()

    def pause(self):
        self._player.pause()

    def play_pause(self):
        self._player.play_pause()

    def stop(self):
        self._player.destroy()

    def set_position(self, track_id: str, position: int):
        self._player.position = position / US_TO_SEC

    def seek(self, offset: int):
        self._player.position = self._player.position / NS_TO_SEC + offset / US_TO_SEC

    def get(self, interface: str, property_name: str) -> GLib.Variant:
        if property_name in {"CanQuit", "CanControl"}:
            return GLib.Variant("b", True)
        elif property_name in {"CanRaise", "HasTrackList"}:
            return GLib.Variant("b", False)
        elif property_name in {
            "CanGoNext",
            "CanGoPrevious",
            "CanPlay",
            "CanPause",
            "CanSeek",
        }:
            return GLib.Variant("b", self._player.loaded_book is not None)
        elif property_name in {"SupportedUriSchemes", "SupportedMimeTypes"}:
            return GLib.Variant("as", [])

        # Might raise an AttributeError. We handle that in Server.on_method_call
        return getattr(self, to_snake_case(property_name))

    def get_all(self, interface):
        if interface == self.MEDIA_PLAYER2_INTERFACE:
            properties = (
                "CanQuit",
                "CanRaise",
                "HasTrackList",
                "Identity",
                "DesktopEntry",
                "SupportedUriSchemes",
                "SupportedMimeTypes",
            )
        elif interface == self.MEDIA_PLAYER2_PLAYER_INTERFACE:
            properties = (
                "PlaybackStatus",
                "Metadata",
                "Position",
                "CanGoNext",
                "CanGoPrevious",
                "CanPlay",
                "CanPause",
                "CanSeek",
                "CanControl",
                "Volume",
            )

        return {property: self.get(interface, property) for property in properties}

    def properties_changed(self, iface_name, changed_props, invalidated_props):
        self._bus.emit_signal(
            None,
            "/org/mpris/MediaPlayer2",
            "org.freedesktop.DBus.Properties",
            "PropertiesChanged",
            GLib.Variant.new_tuple(
                GLib.Variant("s", iface_name),
                GLib.Variant("a{sv}", changed_props),
                GLib.Variant("as", invalidated_props),
            ),
        )

    @property
    def desktop_entry(self):
        return GLib.Variant("s", "com.github.geigi.cozy")

    @property
    def identity(self):
        return GLib.Variant("s", "Cozy")

    @property
    def playback_status(self):
        if self._player.playing:
            return GLib.Variant("s", "Playing")
        elif not self._player.loaded_book:
            return GLib.Variant("s", "Stopped")
        else:
            return GLib.Variant("s", "Paused")

    @property
    def metadata(self):
        return GLib.Variant("a{sv}", self._metadata)

    @property
    def position(self):
        return GLib.Variant("x", round(self._player.position / 1e3))

    @property
    def volume(self):
        return GLib.Variant("d", self._player.volume)

    def _get_track_id(self) -> float:
        """
        Track IDs must be unique even up to the point that if a song
        is repeated in a playlist it must have a different TrackId.
        """
        return time.time() * 1e10 % 1e10

    def _get_new_metadata(self, book: Book | None = None) -> dict[str, GLib.Variant]:
        if book is None:
            return Metadata.no_track()

        track_path_template = "/com/github/geigi/cozy/TrackId/{id:.0f}"
        uri_template = "file://{path}"

        metadata = Metadata(
            track_id=track_path_template.format(id=self._get_track_id()),
            track_number=book.current_chapter.number,
            title=book.current_chapter.name,
            album=book.name,
            artist=[book.author],
            length=book.current_chapter.length * US_TO_SEC,
            url=uri_template.format(path=book.current_chapter.file),
            artwork_uri=self._artwork_cache.get_album_art_path(book, 256),
        )
        return metadata.to_dict()

    def _on_player_changed(self, event: str, _) -> None:
        if event == "chapter-changed":
            self._on_current_changed()
        elif event == "play":
            self._on_status_changed("Playing")
        elif event == "pause":
            self._on_status_changed("Paused")
        elif event == "stop":
            self._on_status_changed("Stopped")

    def _on_app_setting_changed(self, event: str, _):
        if event == "swap-author-reader":
            self._on_current_changed()

    def _on_current_changed(self) -> None:
        if not self._player.loaded_book:
            return

        self._metadata = self._get_new_metadata(self._player.loaded_book)

        properties = {
            "Metadata": GLib.Variant("a{sv}", self._metadata),
            "CanPlay": GLib.Variant("b", True),
            "CanPause": GLib.Variant("b", True),
            "CanGoNext": GLib.Variant("b", True),
            "CanGoPrevious": GLib.Variant("b", True),
        }

        self.properties_changed(self.MEDIA_PLAYER2_PLAYER_INTERFACE, properties, [])

    def _on_status_changed(self, status: str) -> None:
        properties = {"PlaybackStatus": GLib.Variant("s", status)}
        self.properties_changed(self.MEDIA_PLAYER2_PLAYER_INTERFACE, properties, [])



import logging
from threading import Thread

from peewee import SqliteDatabase

from cozy.application_settings import ApplicationSettings
from cozy.architecture.event_sender import EventSender
from cozy.architecture.observable import Observable
from cozy.control.filesystem_monitor import FilesystemMonitor
from cozy.ext import inject
from cozy.media.importer import Importer
from cozy.model.library import Library
from cozy.model.settings import Settings
from cozy.model.storage import Storage

log = logging.getLogger("storages_view_model")


class StoragesViewModel(Observable, EventSender):
    _library: Library = inject.attr(Library)
    _importer: Importer = inject.attr(Importer)
    _model: Settings = inject.attr(Settings)
    _app_settings: ApplicationSettings = inject.attr(ApplicationSettings)
    _db = inject.attr(SqliteDatabase)
    _fs_monitor = inject.attr(FilesystemMonitor)

    def __init__(self) -> None:
        super().__init__()
        super(Observable, self).__init__()

        self._selected_storage = None

    def _scan_new_storage(self, model: Storage) -> None:
        self.emit_event("storage-added", model)
        log.info("New audiobook location added. Starting import scan.")
        Thread(target=self._importer.scan, name="ImportThread").start()

    def _rebase_storage_location(self, model: Storage, old_path: str) -> None:
        self.emit_event("storage-changed", model)
        log.info("Audiobook location changed, rebasing the location in Cozy.")
        Thread(
            target=self._library.rebase_path,
            args=(old_path, model.path),
            name="RebaseStorageLocationThread",
        ).start()

    def add_storage_location(self, path: str | None) -> None:
        if path is None:
            return

        model = Storage.new(self._db, path)
        model.external = self._fs_monitor.is_external(path)

        self._model.invalidate()
        self._notify("storage_locations")

        self._scan_new_storage(model)

    def add_first_storage_location(self, path: str) -> None:
        storage = self.storages[0]
        storage.path = path
        storage.external = self._fs_monitor.is_external(path)
        assert storage.default

        self._model.invalidate()
        self._notify("storage_locations")

    def change_storage_location(self, model: Storage, new_path: str) -> None:
        old_path = model.path
        model.path = new_path
        model.external = self._fs_monitor.is_external(new_path)

        self._rebase_storage_location(model, old_path)
        self._notify("storage_attributes")

    @property
    def storages(self) -> list[Storage]:
        return self._model.storage_locations

    @property
    def default(self) -> Storage | None:
        for item in self.storages:
            if item.default:
                return item

    @property
    def selected_storage(self) -> Storage | None:
        return self._selected_storage

    @selected_storage.setter
    def selected_storage(self, value) -> None:
        self._selected_storage = value

    def remove(self, model: Storage) -> None:
        if model.default:
            return

        model.delete()
        self._model.invalidate()

        storage_path = str(model.path)
        for book in self._library.books:
            chapters_to_remove = [
                c for c in book.chapters if c.file.startswith(storage_path)
            ]

            for chapter in chapters_to_remove:
                chapter.delete()

        self.emit_event("storage-removed", model)
        self._notify("storage_locations")

    def set_default(self, model: Storage) -> None:
        if model.default:
            return

        for storage in self.storages:
            storage.default = False

        model.default = True

        self._notify("storage_attributes")

    def set_external(self, model: Storage, external: bool) -> None:
        model.external = external

        if external:
            self.emit_event("external-storage-added", model)
        else:
            self.emit_event("external-storage-removed", model)

        self._notify("storage_attributes")



from typing import Callable

from gi.repository import Adw, Gtk


class ArtistResultRow(Adw.ActionRow):
    def __init__(self, name: str, on_click: Callable[[str], None]) -> None:
        super().__init__(
            title=name,
            selectable=False,
            activatable=True,
            use_markup=False,
            tooltip_text=_("Jump to {artist_name}").format(artist_name=name),
        )

        self.connect("activated", lambda *_: on_click(name))

        icon = Gtk.Image.new_from_icon_name("person-symbolic")
        icon.set_pixel_size(24)
        icon.set_margin_top(6)
        icon.set_margin_bottom(6)

        self.add_prefix(icon)



from gi.repository import Adw, Gtk

from cozy.ext import inject
from cozy.model.book import Book
from cozy.ui.widgets.book_row import BookRow


class DeleteBookView(Adw.MessageDialog):
    main_window = inject.attr("MainWindow")

    def __init__(self, callback, book: Book):
        super().__init__(
            heading=_("Delete Audiobook?"),
            body=_("The audiobook will be removed from your disk and from Cozy's library."),
            default_response="cancel",
            close_response="cancel",
            transient_for=self.main_window.window,
            modal=True,
        )

        self.add_response("cancel", _("Cancel"))
        self.add_response("delete", _("Remove Audiobook"))
        self.set_response_appearance("delete", Adw.ResponseAppearance.DESTRUCTIVE)

        list_box = Gtk.ListBox(margin_top=12, css_classes=["boxed-list"])
        list_box.append(BookRow(book))
        self.set_extra_child(list_box)

        self.connect("response", callback, book)



from typing import Callable

from gi.repository import Adw, Gio, GLib, GObject, Gtk

from cozy.ext import inject
from cozy.model.storage import Storage
from cozy.view_model.storages_view_model import StoragesViewModel


def ask_storage_location(callback: Callable[[str | None], None], initial_folder: str | None = None):
    location_chooser = Gtk.FileDialog(title=_("Set Audiobooks Directory"))

    if initial_folder:
        gfile = Gio.File.new_for_path(initial_folder)
        location_chooser.set_initial_folder(gfile)

    def finish_callback(dialog, result):
        try:
            file = dialog.select_folder_finish(result)
        except GLib.GError:
            pass
        else:
            callback(None if file is None else file.get_path())

    location_chooser.select_folder(inject.instance("MainWindow").window, None, finish_callback)


@Gtk.Template.from_resource("/com/github/geigi/cozy/storage_row.ui")
class StorageRow(Adw.ActionRow):
    __gtype_name__ = "StorageRow"

    icon: Gtk.Image = Gtk.Template.Child()
    default_icon: Gtk.Image = Gtk.Template.Child()
    menu_button: Gtk.MenuButton = Gtk.Template.Child()

    def __init__(self, model: Storage, menu_model: Gio.Menu) -> None:
        self._model = model

        super().__init__(title=model.path)
        self.connect("activated", self.ask_for_new_location)

        self.menu_button.set_menu_model(menu_model)
        self.menu_button.connect("notify::active", self._on_menu_opened)

        self._set_default_icon()
        self._set_drive_icon()

    @property
    def model(self) -> Storage:
        return self._model

    def ask_for_new_location(self, *_) -> None:
        ask_storage_location(self._on_folder_changed, initial_folder=self._model.path)

    def _on_folder_changed(self, new_path: str | None) -> None:
        if new_path is not None:
            self.emit("location-changed", new_path)

    def _on_menu_opened(self, *_) -> None:
        self.emit("menu-opened")

    def _set_drive_icon(self) -> None:
        if self._model.external:
            self.icon.set_from_icon_name("network-server-symbolic")
            self.icon.set_tooltip_text(_("External drive"))
        else:
            self.icon.set_from_icon_name("folder-open-symbolic")
            self.icon.set_tooltip_text(_("Internal drive"))

    def _set_default_icon(self) -> None:
        self.default_icon.set_visible(self._model.default)


GObject.signal_new(
    "location-changed",
    StorageRow,
    GObject.SIGNAL_RUN_LAST,
    GObject.TYPE_PYOBJECT,
    (GObject.TYPE_PYOBJECT,),
)
GObject.signal_new("menu-opened", StorageRow, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT, ())


@Gtk.Template.from_resource("/com/github/geigi/cozy/storage_locations.ui")
class StorageLocations(Adw.PreferencesGroup):
    __gtype_name__ = "StorageLocations"

    _view_model: StoragesViewModel = inject.attr(StoragesViewModel)

    storage_locations_list: Gtk.ListBox = Gtk.Template.Child()
    storage_menu: Gio.Menu = Gtk.Template.Child()

    def __init__(self) -> None:
        super().__init__()

        self._view_model.bind_to("storage_locations", self._reload_storage_list)
        self._view_model.bind_to("storage_attributes", self._reload_storage_list)

        self._create_actions()
        self.new_storage_button = self._create_new_storage_button()

        self._reload_storage_list()

    def _create_actions(self) -> None:
        self.action_group = Gio.SimpleActionGroup.new()
        self.insert_action_group("storage", self.action_group)

        self.set_external_action = Gio.SimpleAction.new_stateful(
            "mark-external", None, GLib.Variant.new_boolean(False)
        )
        self._set_external_signal_handler = self.set_external_action.connect(
            "notify::state", self._mark_storage_location_external
        )
        self.action_group.add_action(self.set_external_action)

        self.remove_action = Gio.SimpleAction.new("remove", None)
        self.remove_action.connect("activate", self._remove_storage_location)
        self.action_group.add_action(self.remove_action)

        self.make_default_action = Gio.SimpleAction.new("make-default", None)
        self.make_default_action.connect("activate", self._set_default_storage_location)
        self.action_group.add_action(self.make_default_action)

    def _create_new_storage_button(self) -> Adw.ActionRow:
        icon = Gtk.Image(icon_name="list-add-symbolic", margin_top=18, margin_bottom=18)
        row = Adw.ActionRow(selectable=False, activatable=True)
        row.connect("activated", self._on_new_storage_clicked)
        row.set_child(icon)
        return row

    def _reload_storage_list(self) -> None:
        self.storage_locations_list.remove_all()

        for storage in self._view_model.storages:
            row = StorageRow(storage, menu_model=self.storage_menu)
            row.connect("location-changed", self._on_storage_location_changed)
            row.connect("menu-opened", self._on_storage_menu_opened)
            self.storage_locations_list.append(row)

        self.storage_locations_list.append(self.new_storage_button)

    def _remove_storage_location(self, *_) -> None:
        self._view_model.remove(self._view_model.selected_storage)

    def _set_default_storage_location(self, *_) -> None:
        self._view_model.set_default(self._view_model.selected_storage)

    def _mark_storage_location_external(
        self, action: Gio.SimpleAction, value: GObject.ParamSpec
    ) -> None:
        value = action.get_property(value.name)
        self._view_model.set_external(self._view_model.selected_storage, value)

    def _on_new_storage_clicked(self, *_) -> None:
        ask_storage_location(self._view_model.add_storage_location)

    def _on_storage_location_changed(self, widget: StorageRow, new_location: str) -> None:
        self._view_model.change_storage_location(widget.model, new_location)

    def _on_storage_menu_opened(self, widget: StorageRow) -> None:
        with self.set_external_action.handler_block(self._set_external_signal_handler):
            self.set_external_action.props.state = GLib.Variant.new_boolean(widget.model.external)

        self.remove_action.props.enabled = (
            not widget.model.default and len(self._view_model.storages) > 1
        )
        self.make_default_action.props.enabled = widget.model is not self._view_model.default
        self._view_model.selected_storage = widget.model



from pathlib import Path

from gi.repository import Adw, Gio, GLib, Gtk

from cozy.ext import inject
from cozy.media.importer import Importer
from cozy.model.chapter import Chapter


class FileNotFoundDialog(Adw.MessageDialog):
    main_window = inject.attr("MainWindow")
    _importer: Importer = inject.attr(Importer)

    def __init__(self, chapter: Chapter):
        self.missing_chapter = chapter

        super().__init__(
            heading=_("File not found"),
            body=_("This file could not be found. Do you want to locate it manually?"),
            default_response="locate",
            close_response="cancel",
            transient_for=self.main_window.window,
            modal=True,
        )

        self.add_response("cancel", _("Cancel"))
        self.add_response("locate", _("Locate"))
        self.set_response_appearance("locate", Adw.ResponseAppearance.SUGGESTED)

        label = Gtk.Label(label=chapter.file, margin_top=12)
        label.add_css_class("monospace")
        self.set_extra_child(label)

        self.connect("response", self._on_locate)

    def _on_locate(self, __, response):
        if response == "locate":
            file_dialog = Gtk.FileDialog(title=_("Locate Missing File"))

            extension = Path(self.missing_chapter.file).suffix[1:]
            current_extension_filter = Gtk.FileFilter(name=_("{ext} files").format(ext=extension))
            current_extension_filter.add_suffix(extension)

            audio_files_filter = Gtk.FileFilter(name=_("Audio files"))
            audio_files_filter.add_mime_type("audio/*")

            filters = Gio.ListStore.new(Gtk.FileFilter)
            filters.append(current_extension_filter)
            filters.append(audio_files_filter)

            file_dialog.set_filters(filters)
            file_dialog.set_default_filter(current_extension_filter)
            file_dialog.open(self.main_window.window, None, self._file_dialog_open_callback)

    def _file_dialog_open_callback(self, dialog, result):
        try:
            file = dialog.open_finish(result)
        except GLib.GError:
            pass
        else:
            if file is not None:
                self.missing_chapter.file = file.get_path()
                self._importer.scan()




import logging
import os
from collections import defaultdict
from threading import Thread
from typing import Callable

from gi.repository import Adw, Gdk, Gio, GLib, Gtk

import cozy.control.filesystem_monitor as fs_monitor
import cozy.ext.inject as inject
import cozy.report.reporter as report
from cozy.application_settings import ApplicationSettings
from cozy.architecture.event_sender import EventSender
from cozy.architecture.singleton import Singleton
from cozy.control.db import books, close_db
from cozy.media.files import Files
from cozy.media.importer import Importer, ScanStatus
from cozy.media.player import Player
from cozy.model.settings import Settings as SettingsModel
from cozy.view_model.settings_view_model import SettingsViewModel
from cozy.view_model.storages_view_model import StoragesViewModel
from cozy.open_view import OpenView
from cozy.ui.library_view import LibraryView
from cozy.ui.preferences_view import PreferencesView
from cozy.ui.widgets.first_import_button import FirstImportButton


log = logging.getLogger("ui")


class CozyUI(EventSender, metaclass=Singleton):
    """
    CozyUI is the main ui class.
    """
    # Is currently an dialog open?
    is_initialized = False
    __inhibit_cookie = None
    fs_monitor = inject.attr(fs_monitor.FilesystemMonitor)
    application_settings = inject.attr(ApplicationSettings)
    _importer: Importer = inject.attr(Importer)
    _settings: SettingsModel = inject.attr(SettingsModel)
    _files: Files = inject.attr(Files)
    _player: Player = inject.attr(Player)
    _storages_view_model: StoragesViewModel = inject.attr(StoragesViewModel)

    def __init__(self, pkgdatadir, app, version):
        super().__init__()
        self.pkgdir = pkgdatadir
        self.app = app
        self.version = version

        self._library_view: LibraryView = None

    def activate(self, library_view: LibraryView):
        self.__init_window()
        self.__init_actions()
        self.__init_components()

        self._library_view = library_view

        self.auto_import()
        self.check_for_tracks()

        self.is_initialized = True

    def startup(self):
        self.__init_resources()

    def __init_resources(self):
        """
        Initialize all resources like gresource and glade windows.
        """

        self.appdata_resource = Gio.resource_load(
            os.path.join(self.pkgdir, 'com.github.geigi.cozy.appdata.gresource'))
        Gio.Resource._register(self.appdata_resource)

        self.window_builder = Gtk.Builder.new_from_resource(
            "/com/github/geigi/cozy/main_window.ui")

        self.window: Gtk.Window = self.window_builder.get_object("app_window")

    def __init_window(self):
        """
        Add fields for all UI objects we need to access from code.
        Initialize everything we can't do from the UI files like events and other stuff.
        """
        log.info("Initializing main window")
        self._restore_window_size()
        self.window.set_title("Cozy")
        self.window.set_application(self.app)

        self.window.connect("close-request", self.on_close)
        self.window.connect("notify::default-width", self._on_window_size_allocate)
        self.window.connect("notify::default-height", self._on_window_size_allocate)

        self._drop_target = Gtk.DropTarget()
        self._drop_target.set_gtypes([Gdk.FileList])
        self._drop_target.set_actions(Gdk.DragAction.COPY)
        self._drop_target.connect("enter", self._on_drag_enter)
        self._drop_target.connect("leave", self._on_drag_leave)
        self._drop_target.connect("drop", self._on_drag_data_received)
        self.window.add_controller(self._drop_target)

        self.main_stack: Gtk.Stack = self.window_builder.get_object("main_stack")
        self.navigation_view: Adw.NavigationView = self.window_builder.get_object("navigation_view")
        self.drop_revealer: Gtk.Revealer = self.window_builder.get_object("drop_revealer")

        self.window.present()

    def __init_actions(self):
        """
        Init all app actions.
        """

        about_action = Gio.SimpleAction.new("about", None)
        about_action.connect("activate", self.about)
        self.app.add_action(about_action)
        self.app.set_accels_for_action("app.about", ["F1"])

        self.create_action("about", self.about)
        self.create_action("quit", self.quit, ["<primary>q", "<primary>w"])
        self.create_action("prefs", self.show_prefs, ["<primary>comma"])
        self.scan_action = self.create_action("scan", self.scan)
        self.play_pause_action = self.create_action("play_pause", self.play_pause, ["space"])

        self.hide_offline_action = Gio.SimpleAction.new_stateful(
            "hide_offline", None, GLib.Variant.new_boolean(self.application_settings.hide_offline)
        )
        self.hide_offline_action.connect("change-state", self.__on_hide_offline)
        self.app.add_action(self.hide_offline_action)

    def create_action(
        self,
        name: str,
        callback: Callable[[Gio.SimpleAction, None], None],
        shortcuts: list[str] | None = None,
    ) -> Gio.SimpleAction:
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.app.add_action(action)

        if shortcuts:
            self.app.set_accels_for_action(f"app.{name}", shortcuts)

        return action

    def __init_components(self):
        path = self._settings.default_location.path if self._settings.storage_locations else None
        self.import_button = FirstImportButton(self._set_audiobook_path, path)
        self.get_object("welcome_status_page").set_child(self.import_button)

        if not self._player.loaded_book:
            self.block_ui_buttons(True)

        self._importer.add_listener(self._on_importer_event)

    def get_object(self, name):
        return self.window_builder.get_object(name)

    def quit(self, action, parameter):
        """
        Quit app.
        """
        self.on_close(None)
        self.app.quit()

    def _get_contributors(self):
        authors_file = self.appdata_resource.lookup_data("/com/github/geigi/cozy/authors", Gio.ResourceLookupFlags.NONE)

        current_section = ""
        result = defaultdict(list)
        for line in authors_file.get_data().decode().splitlines():
            if line.startswith("#"):
                current_section = line[1:].strip().lower()
            elif line.startswith("-"):
                result[current_section].append(line[1:].strip())

        return result

    def about(self, *junk):
        """
        Show about window.
        """
        about = Adw.AboutWindow.new_from_appdata(
            "/com/github/geigi/cozy/com.github.geigi.cozy.appdata.xml",
            release_notes_version=self.version,
        )

        contributors = self._get_contributors()
        about.set_developers(sorted(contributors["code"]))
        about.set_designers(sorted(contributors["design"]))
        about.set_artists(sorted(contributors["icon"]))

        about.set_license_type(Gtk.License.GPL_3_0)

        about.add_acknowledgement_section(
            _("Patreon Supporters"),
            ["Fred Warren", "Gabriel", "Hu Mann", "Josiah", "Oleksii Kriukov"]
        )
        about.add_acknowledgement_section(
            _("m4b chapter support in mutagen"),
            ("mweinelt",),
        )
        about.add_acknowledgement_section(
            _("Open Source Projects"),
            ("Lollypop music player https://gitlab.gnome.org/World/lollypop",),
        )

        # Translators: Replace "translator-credits" with your names, one name per line
        about.set_translator_credits(_("translator-credits"))
        about.add_legal_section("python-inject", "© 2010 Ivan Korobkov", Gtk.License.APACHE_2_0)

        about.set_transient_for(self.window)
        about.present()

    def show_prefs(self, *_):
        """
        Show preferences window.
        """
        PreferencesView().present()

    def play_pause(self, *_):
        self._player.play_pause()

    def block_ui_buttons(self, block, scan=False):
        """
        Makes the buttons to interact with the player insensetive.
        :param block: Boolean
        """
        sensitive = not block
        try:
            self.play_pause_action.set_enabled(sensitive)
            if scan:
                self.scan_action.set_enabled(sensitive)
                self.hide_offline_action.set_enabled(sensitive)
        except GLib.GError:
            pass

    def switch_to_playing(self):
        """
        Switch the UI state back to playing.
        This enables all UI functionality for the user.
        """
        if self.navigation_view.props.visible_page != "book_overview" and self.main_stack.props.visible_child_name != "welcome":
            self.navigation_view.pop_to_tag("main")

        if self._player.loaded_book:
            self.block_ui_buttons(False, True)
        else:
            # we want to only block the player controls
            # TODO: rework. this is messy
            self.block_ui_buttons(False, True)
            self.block_ui_buttons(True, False)

    def check_for_tracks(self):
        """
        Check if there are any imported files.
        If there aren't display a welcome screen.
        """
        if books().count() < 1:
            self.block_ui_buttons(True)

    def scan(self, _, __):
        thread = Thread(target=self._importer.scan, name="ScanMediaThread")
        thread.start()

    def auto_import(self):
        if self.application_settings.autoscan:
            self.scan(None, None)

    def __on_hide_offline(self, action, value):
        """
        Show/Hide offline books action handler.
        """
        action.set_state(value)
        self.application_settings.hide_offline = value.get_boolean()

    def _on_drag_enter(self, *_):
        self.drop_revealer.set_reveal_child(True)
        self.main_stack.add_css_class("blurred")
        return True

    def _on_drag_leave(self, *_):
        self.drop_revealer.set_reveal_child(False)
        self.main_stack.remove_css_class("blurred")
        return True

    def _on_drag_data_received(self, widget, value, *_):
        thread = Thread(target=self._files.copy, args=[value.get_files()], name="DnDImportThread")
        thread.start()
        return True

    def _set_audiobook_path(self, path: str | None) -> None:
        if path is None:
            return

        self.import_button.disable()
        self._storages_view_model.add_first_storage_location(path)
        self.scan(None, None)
        self.fs_monitor.init_offline_mode()

    def on_close(self, widget, data=None):
        """
        Close and dispose everything that needs to be when window is closed.
        """
        log.info("Closing.")
        self.fs_monitor.close()

        self._player.destroy()

        close_db()

        report.close()

        log.info("Closing app.")
        self.app.quit()
        log.info("App closed.")

    def get_builder(self):
        return self.window_builder

    def _on_importer_event(self, event: str, message):
        if event == "scan" and message == ScanStatus.SUCCESS:
            self.check_for_tracks()

    def _restore_window_size(self):
        width = self.application_settings.window_width
        height = self.application_settings.window_height
        self.window.set_default_size(width, height)
        if self.application_settings.window_maximize:
            self.window.maximize()
        else:
            self.window.unmaximize()

    def _on_window_size_allocate(self, *_):
        width, height = self.window.get_default_size()
        self.application_settings.window_width = width
        self.application_settings.window_height = height
        self.application_settings.window_maximize = self.window.is_maximized()




from gi.repository import Gdk, GObject, Gtk

from cozy.control.string_representation import seconds_to_str


@Gtk.Template.from_resource('/com/github/geigi/cozy/seek_bar.ui')
class SeekBar(Gtk.Box):
    __gtype_name__ = "SeekBar"

    progress_scale: Gtk.Scale = Gtk.Template.Child()
    current_label: Gtk.Label = Gtk.Template.Child()
    remaining_label: Gtk.Label = Gtk.Template.Child()
    remaining_event_box: Gtk.Box = Gtk.Template.Child()

    length: float

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.length: float = 0.0
        self._progress_scale_pressed = False

        self.progress_scale.connect("value-changed", self._on_progress_scale_changed)

        # HACK: Using a GtkGestureClick here is not possible, as GtkRange's internal
        # gesture controller claims the button press event, and thus the released signal doesn't get emitted.
        # Therefore we get its internal GtkGestureClick, and add our handlers to that.
        # Hacky workaround from: https://gitlab.gnome.org/GNOME/gtk/-/issues/4939
        # Ideally GtkRange would forward these signals, so we wouldn't need this hack
        # TODO: Add these signals to Gtk and make a MR?
        for controller in self.progress_scale.observe_controllers():
            if isinstance(controller, Gtk.GestureClick):
                click_gesture = controller
                break

        click_gesture.set_button(0)  # Enable all mouse buttons
        click_gesture.connect("pressed", self._on_progress_scale_press)
        click_gesture.connect("released", self._on_progress_scale_release)

        keyboard_controller = Gtk.EventControllerKey()
        keyboard_controller.connect("key-pressed", self._on_progress_key_pressed)
        self.progress_scale.add_controller(keyboard_controller)

    @property
    def position(self) -> float:
        return self.progress_scale.get_value()

    @position.setter
    def position(self, new_value: float):
        if not self._progress_scale_pressed:
            self.progress_scale.set_value(new_value)

    @property
    def sensitive(self) -> bool:
        return self.progress_scale.get_sensitive()

    @sensitive.setter
    def sensitive(self, new_value: bool):
        self.progress_scale.set_sensitive(new_value)

    @property
    def visible(self) -> bool:
        return self.progress_scale.get_visible()

    @visible.setter
    def visible(self, value: bool):
        self.current_label.set_visible(value)
        self.progress_scale.set_visible(value)
        self.remaining_event_box.set_visible(value)

    def _on_progress_scale_changed(self, _):
        total = self.length
        position = int(total * self.progress_scale.get_value() / 100)
        remaining_secs = int(total - position)

        self.current_label.set_text(seconds_to_str(position, total))
        self.remaining_label.set_text(seconds_to_str(remaining_secs, total))

    def _on_progress_scale_release(self, *_):
        self._progress_scale_pressed = False
        value = self.progress_scale.get_value()
        self.emit("position-changed", value)

    def _on_progress_key_pressed(self, _, event, *__):
        if event in {Gdk.KEY_Up, Gdk.KEY_Left}:
            self.emit("rewind")
        elif event in {Gdk.KEY_Down, Gdk.KEY_Right}:
            self.emit("forward")

    def _on_progress_scale_press(self, *_):
        self._progress_scale_pressed = True


GObject.signal_new('position-changed', SeekBar, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT,
                   (GObject.TYPE_PYOBJECT,))

GObject.signal_new('rewind', SeekBar, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT, ())
GObject.signal_new('forward', SeekBar, GObject.SIGNAL_RUN_LAST, GObject.TYPE_PYOBJECT, ())
